#include <cmath>
#include <iostream>
#include <GL/glew.h>
#include <GL/freeglut.h>
#include <GL/freeglut_ext.h>
#include <glm/glm.hpp>
#include <vector>
#include "moglclass.h"
#include "shaders.h"
#include "vertexBufferObject.h"

using namespace moglclass;

#define NUMSHADERS 2
Shader shShaders[NUMSHADERS];
Program spMain;

// 2D gasket point
GLuint gasket2DPointVAO;
VertexBufferObject gasket2DPointVBO;
const int gasket2DPointNumVertices = 5000;
glm::vec4 gasket2DPointVertices[3] = { glm::vec4(-1.0, -1.0, 0.0, 1.0), glm::vec4(0.0, 1.0, 0.0, 1.0), glm::vec4(1.0, -1.0, 0.0, 1.0) };
glm::vec4 gasket2DPoints[gasket2DPointNumVertices];

// 2D gasket triangle
GLuint gasket2DTriangleVAO;
VertexBufferObject gasket2DTriangleVBO;
const int gasket2DTriangleNumTimesToSubdivide = 5;
const int gasket2DTriangleNumTriangles = 243;  // 3^5 triangles generated
const int gasket2DTriangleNumVertices = 3 * gasket2DTriangleNumTriangles;
glm::vec4 gasket2DTrianglePoints[gasket2DTriangleNumVertices];



int g_mode = 0; // 0=2D Sierpinski Gasket Point, 1=2D Sierpinski Gasket Triangle, 2=3D Sierpinski Gasket
GLenum g_primitivemode = GL_POINTS;

////////////////////////////////////////////////////////////////////////////////////////////////////
// set color 2D gasket point
void setDataGasket2DPoint()
{
	// VAO
	glGenVertexArrays(1, &gasket2DPointVAO);
	glBindVertexArray(gasket2DPointVAO);

	// VBO
	gasket2DPointVBO.createVBO();
	gasket2DPointVBO.bindVBO();

	// select an arbitrary initial point inside of the triangle
	gasket2DPoints[0] = glm::vec4(0.25, 0.50, 0.0, 1.0);
	// compute and store N-1 new points
	for (int i = 1; i < gasket2DPointNumVertices; ++i)
	{
		int j = rand() % 3;   // pick a vertex at random
		gasket2DPoints[i] = (gasket2DPoints[i - 1] + gasket2DPointVertices[j]) / 2.0f; // Compute the point halfway between the selected vertex and the previous point
	}

	// set VBO
	for (int i = 0; i < gasket2DPointNumVertices; ++i)
	{
		gasket2DPointVBO.addData(&gasket2DPoints[i], sizeof(glm::vec4)); // vertex position
		gasket2DPointVBO.addData(&glm::vec4(0.0f, 0.0f, 1.0f, 1.0f), sizeof(glm::vec4)); // vertex color
	}
	gasket2DPointVBO.uploadDataToGPU(GL_STATIC_DRAW);

	// vertex attrib
	int iDataStride = 2 * sizeof(glm::vec4); // vertex & color
	int iDataOffset = 0;
	glEnableVertexAttribArray(0);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
	iDataOffset += sizeof(glm::vec4);
	glEnableVertexAttribArray(1);
	glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
}

void init( void )
{
    // load shaders and use the resulting shader program
	spMain.createProgram();
	shShaders[0].loadShader("simple2.vs", GL_VERTEX_SHADER);
	shShaders[1].loadShader("simple2.fs", GL_FRAGMENT_SHADER);
	spMain.attachShaderToProgram(&shShaders[0]);
	spMain.attachShaderToProgram(&shShaders[1]);
	spMain.linkProgram();
	spMain.useProgram();
	
	// set data
	setDataGasket2DPoint();

	// set white background
	glClearColor( 1.0, 1.0, 1.0, 1.0 ); 
}

void drawGasket2DPoint()
{
	glBindVertexArray(gasket2DPointVAO);
	glDrawArrays(GL_POINTS, 0, gasket2DPointNumVertices);       // draw 2D gasket points
	glBindVertexArray(0);
	glDisableVertexAttribArray(0);
}


void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT );     // clear the window

	drawGasket2DPoint(); // draw gasket

	glFlush();
}

void keyboard( unsigned char key, int x, int y )
{
	switch (key)  
	{    
	case 0x1B:    
	case 'q':    
	case 'Q':      
		exit(0);     
	case 'm':
		g_mode = (g_mode + 1) % 3;
		printf("g_mode=%d\n", g_mode);
		break;
	}
	glutPostRedisplay();
}

void reshape(int w, int h)
{
	glViewport(0, 0, w, h);
}

void cleanup()
{
	spMain.deleteProgram();
}

int main( int argc, char **argv )
{
	glutInit(&argc, argv);
    glutInitDisplayMode( GLUT_RGBA );
    glutInitWindowSize( 512, 512 );
    glutCreateWindow(argv[0]);
    
	glewInit();
    init();
    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
	glutReshapeFunc( reshape );
	glutMainLoop();

	cleanup();
    return 0;
}
