#include <cmath>
#include <iostream>
#include <GL/glew.h>
#include <GL/freeglut.h>
#include <GL/freeglut_ext.h>
#include <glm/glm.hpp>
#include <vector>
#include "moglclass.h"
#include "shaders.h"
#include "vertexBufferObject.h"

using namespace moglclass;

#ifndef M_PI
#define M_PI 3.141592
#endif
#define MOGL_STEP 7
#define MOGL_THICKNESS 3

#define NUMSHADERS 2
Shader shShaders[NUMSHADERS];
Program spMain;

// Indices into vertices[] buffer
std::vector<int> indices; // GL_LINES, GL_LINE_STRIP, GL_TRIANGLES, GL_TRIANGLE_STRIP

std::vector<glm::vec4> vertices;
GLuint vao[4];
GLuint vbo[4];
GLuint ibo[4];
int g_count[4];
GLuint program;

int g_colormode = 0;
int g_primitivemode = 0;
bool g_wiremode = false;

////////////////////////////////////////////////////////////////////////////////////////////////////

void SetData1() // GL_LINES
{
	int i = 0;

	// position
	vertices.push_back(glm::vec4(0.5f, 0.25f, 0.0f, 1.0f)); // 0
	vertices.push_back(glm::vec4(0.5f, -0.25f, 0.0f, 1.0f)); // 1
	vertices.push_back(glm::vec4(0.25f, 0.25f, 0.0f, 1.0f)); // 2
	vertices.push_back(glm::vec4(0.25f, -0.25f, 0.0f, 1.0f)); // 3
	vertices.push_back(glm::vec4(0, 0.25f, 0.0f, 1.0f)); //4
	vertices.push_back(glm::vec4(0, -0.25f, 0.0f, 1.0f)); //5
	vertices.push_back(glm::vec4(-0.25f, 0.25f, 0.0f, 1.0f)); //6
	vertices.push_back(glm::vec4(-0.25f, -0.25f, 0.0f, 1.0f)); //7
	vertices.push_back(glm::vec4(-0.5f, 0.25f, 0.0f, 1.0f)); //8
	vertices.push_back(glm::vec4(-0.5f, -0.25f, 0.0f, 1.0f)); //9
	printf("multiquad::numVertices =%d\n", (int)vertices.size());

	indices.push_back(1);
	indices.push_back(2);
	indices.push_back(2);
	indices.push_back(3);
	indices.push_back(3);
	indices.push_back(4);
	indices.push_back(4);
	indices.push_back(5);
	indices.push_back(5);
	indices.push_back(6);
	indices.push_back(6);
	indices.push_back(7);
	indices.push_back(7);
	indices.push_back(8);
	indices.push_back(8);
	indices.push_back(9);
	indices.push_back(9);
	indices.push_back(7);
	indices.push_back(7);
	indices.push_back(5);
	indices.push_back(5);
	indices.push_back(3);
	indices.push_back(3);
	indices.push_back(1);
	indices.push_back(1);
	indices.push_back(0);
	indices.push_back(0);
	indices.push_back(2);
	indices.push_back(2);
	indices.push_back(4);
	indices.push_back(4);
	indices.push_back(6);
	indices.push_back(6);
	indices.push_back(8);

	printf("g_count[%d] =%d\n", i, (int)indices.size());
	g_count[i] = (int)indices.size();

	// VAO
	glGenVertexArrays(1, &vao[i]);
	glBindVertexArray(vao[i]);

	// VBO
	glGenBuffers(1, &vbo[i]);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
	glBufferData(GL_ARRAY_BUFFER, 10*sizeof(glm::vec4), &vertices[0], GL_STATIC_DRAW);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(0);
	
	// IBO
	glGenBuffers(1, &ibo[i]);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo[i]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, g_count[i] * sizeof(int), &indices[0], GL_STATIC_DRAW);

	glBindVertexArray(0);

	vertices.clear();
	indices.clear();
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void SetData2() // GL_LINE_STRIP
{
	int i = 1;

	// position
	vertices.push_back(glm::vec4(0.5f, 0.25f, 0.0f, 1.0f)); // 0
	vertices.push_back(glm::vec4(0.5f, -0.25f, 0.0f, 1.0f)); // 1
	vertices.push_back(glm::vec4(0.25f, 0.25f, 0.0f, 1.0f)); // 2
	vertices.push_back(glm::vec4(0.25f, -0.25f, 0.0f, 1.0f)); // 3
	vertices.push_back(glm::vec4(0, 0.25f, 0.0f, 1.0f)); //4
	vertices.push_back(glm::vec4(0, -0.25f, 0.0f, 1.0f)); //5
	vertices.push_back(glm::vec4(-0.25f, 0.25f, 0.0f, 1.0f)); //6
	vertices.push_back(glm::vec4(-0.25f, -0.25f, 0.0f, 1.0f)); //7
	vertices.push_back(glm::vec4(-0.5f, 0.25f, 0.0f, 1.0f)); //8
	vertices.push_back(glm::vec4(-0.5f, -0.25f, 0.0f, 1.0f)); //9
	printf("multiquad::numVertices =%d\n", (int)vertices.size());

	indices.push_back(1);
	indices.push_back(2);
	indices.push_back(3);
	indices.push_back(4);
	indices.push_back(5);
	indices.push_back(6);
	indices.push_back(7);
	indices.push_back(8);
	indices.push_back(9);
	indices.push_back(7);
	indices.push_back(5);
	indices.push_back(3);
	indices.push_back(1);
	indices.push_back(0);
	indices.push_back(2);
	indices.push_back(4);
	indices.push_back(6);
	indices.push_back(8);

	printf("g_count[%d] =%d\n", i, (int)indices.size());
	g_count[i] = (int)indices.size();

	// VAO
	glGenVertexArrays(1, &vao[i]);
	glBindVertexArray(vao[i]);

	// VBO
	glGenBuffers(1, &vbo[i]);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[i]);
	glBufferData(GL_ARRAY_BUFFER, 10 * sizeof(glm::vec4), &vertices[0], GL_STATIC_DRAW);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(0);

	// IBO
	glGenBuffers(1, &ibo[i]);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo[i]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, g_count[i] * sizeof(int), &indices[0], GL_STATIC_DRAW);

	glBindVertexArray(0);

	vertices.clear();
	indices.clear();
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void SetData3() // GL_TRIANGLES
{
	int i = 2;

	// position
	vertices.push_back(glm::vec4(0.5f, 0.25f, 0.0f, 1.0f)); // 0
	vertices.push_back(glm::vec4(0.5f, -0.25f, 0.0f, 1.0f)); // 1
	vertices.push_back(glm::vec4(0.25f, 0.25f, 0.0f, 1.0f)); // 2
	vertices.push_back(glm::vec4(0.25f, -0.25f, 0.0f, 1.0f)); // 3
	vertices.push_back(glm::vec4(0, 0.25f, 0.0f, 1.0f)); //4
	vertices.push_back(glm::vec4(0, -0.25f, 0.0f, 1.0f)); //5
	vertices.push_back(glm::vec4(-0.25f, 0.25f, 0.0f, 1.0f)); //6
	vertices.push_back(glm::vec4(-0.25f, -0.25f, 0.0f, 1.0f)); //7
	vertices.push_back(glm::vec4(-0.5f, 0.25f, 0.0f, 1.0f)); //8
	vertices.push_back(glm::vec4(-0.5f, -0.25f, 0.0f, 1.0f)); //9
	printf("multiquad::numVertices =%d\n", (int)vertices.size());

	indices.push_back(0);
	indices.push_back(1);
	indices.push_back(2);

	indices.push_back(1);
	indices.push_back(2);
	indices.push_back(3);

	indices.push_back(2);
	indices.push_back(3);
	indices.push_back(4);

	indices.push_back(3);
	indices.push_back(4);
	indices.push_back(5);

	indices.push_back(4);
	indices.push_back(5);
	indices.push_back(6);

	indices.push_back(5);
	indices.push_back(6);
	indices.push_back(7);

	indices.push_back(6);
	indices.push_back(7);
	indices.push_back(8);

	indices.push_back(7);
	indices.push_back(8);
	indices.push_back(9);

	printf("g_count[%d] =%d\n", i, (int)indices.size());
	g_count[i] = (int)indices.size();

	// VAO
	glGenVertexArrays(1, &vao[i]);
	glBindVertexArray(vao[i]);

	// VBO
	glGenBuffers(1, &vbo[i]);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[i]);
	glBufferData(GL_ARRAY_BUFFER, 10 * sizeof(glm::vec4), &vertices[0], GL_STATIC_DRAW);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(0);

	// IBO
	glGenBuffers(1, &ibo[i]);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo[i]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, g_count[i] * sizeof(int), &indices[0], GL_STATIC_DRAW);

	glBindVertexArray(0);

	vertices.clear();
	indices.clear();
}

////////////////////////////////////////////////////////////////////////////////////////////////////

void SetData4() // GL_TRIANGLE_STRIP
{
	int i = 3;

	// position
	vertices.push_back(glm::vec4(0.5f, 0.25f, 0.0f, 1.0f)); // 0
	vertices.push_back(glm::vec4(0.5f, -0.25f, 0.0f, 1.0f)); // 1
	vertices.push_back(glm::vec4(0.25f, 0.25f, 0.0f, 1.0f)); // 2
	vertices.push_back(glm::vec4(0.25f, -0.25f, 0.0f, 1.0f)); // 3
	vertices.push_back(glm::vec4(0, 0.25f, 0.0f, 1.0f)); //4
	vertices.push_back(glm::vec4(0, -0.25f, 0.0f, 1.0f)); //5
	vertices.push_back(glm::vec4(-0.25f, 0.25f, 0.0f, 1.0f)); //6
	vertices.push_back(glm::vec4(-0.25f, -0.25f, 0.0f, 1.0f)); //7
	vertices.push_back(glm::vec4(-0.5f, 0.25f, 0.0f, 1.0f)); //8
	vertices.push_back(glm::vec4(-0.5f, -0.25f, 0.0f, 1.0f)); //9
	printf("multiquad::numVertices =%d\n", (int)vertices.size());

	indices.push_back(0);
	indices.push_back(1);
	indices.push_back(2);
	indices.push_back(3);
	indices.push_back(4);
	indices.push_back(5);
	indices.push_back(6);
	indices.push_back(7);
	indices.push_back(8);
	indices.push_back(9);

	printf("g_count[%d] =%d\n", i, (int)indices.size());
	g_count[i] = (int)indices.size();

	// VAO
	glGenVertexArrays(1, &vao[i]);
	glBindVertexArray(vao[i]);

	// VBO
	glGenBuffers(1, &vbo[i]);
	glBindBuffer(GL_ARRAY_BUFFER, vbo[i]);
	glBufferData(GL_ARRAY_BUFFER, 10 * sizeof(glm::vec4), &vertices[0], GL_STATIC_DRAW);
	glVertexAttribPointer(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
	glEnableVertexAttribArray(0);

	// IBO
	glGenBuffers(1, &ibo[i]);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo[i]);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, g_count[i] * sizeof(int), &indices[0], GL_STATIC_DRAW);

	glBindVertexArray(0);

	vertices.clear();
	indices.clear();
}

void init( void )
{
    // Load shaders and use the resulting shader program
	spMain.createProgram();
	shShaders[0].loadShader("simple2.vs", GL_VERTEX_SHADER);
	shShaders[1].loadShader("simple2.fs", GL_FRAGMENT_SHADER);
	spMain.attachShaderToProgram(&shShaders[0]);
	spMain.attachShaderToProgram(&shShaders[1]);
	spMain.linkProgram();
	spMain.useProgram();
	
	// Set data
	SetData1();
	SetData2();
	SetData3();
	SetData4();

	glClearColor( 1.0, 1.0, 1.0, 1.0 ); // white background
}

void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT );     // clear the window

	// multi-quad
	if (g_colormode == 0)
		glVertexAttrib3f((GLuint)1, 1, 0, 0); // set constant color attribute (red)
	else if (g_colormode == 1) 
		glVertexAttrib3f((GLuint)1, 0, 1, 0); // set constant color attribute (green)
	else if (g_colormode == 2) 
		glVertexAttrib3f((GLuint)1, 0, 0, 1); // set constant color attribute (blue)

	glPolygonMode(GL_FRONT, GL_FILL);

	// primitive type
	if (g_primitivemode == 0) // 0
	{
		glBindVertexArray(vao[0]);
		glLineWidth(MOGL_THICKNESS);
		glDrawElements(GL_LINES, g_count[0], GL_UNSIGNED_INT, 0);  // draw lines
		glLineWidth(1);
	}
	else if (g_primitivemode == 1) // 1
	{
		glBindVertexArray(vao[1]);
		glLineWidth(MOGL_THICKNESS);
		glDrawElements(GL_LINE_STRIP, g_count[1], GL_UNSIGNED_INT, 0); // draw line strip
		glLineWidth(1);
	}
	else if (g_primitivemode == 2) // 2
	{
		glBindVertexArray(vao[2]);
		glDrawElements(GL_TRIANGLES, g_count[2], GL_UNSIGNED_INT, 0);  // draw triangles
	}
	else if (g_primitivemode == 3) // 3
	{
		glBindVertexArray(vao[3]);
		glDrawElements(GL_TRIANGLE_STRIP, g_count[3], GL_UNSIGNED_INT, 0);     // draw triangle strip
	}

	glBindVertexArray(0);
	glDisableVertexAttribArray(0);

	glFlush();
}

void keyboard( unsigned char key, int x, int y )
{
	switch (key)  
	{    
	case 0x1B:    
	case 'q':    
	case 'Q':      
		exit(0);     
	case ' ':
		g_primitivemode++;
		g_primitivemode = g_primitivemode % 4;
		printf("g_primitivemode=%d\n", g_primitivemode);
		break;
	case 'c':
		g_colormode = (g_colormode+1) % 3;
		printf("g_colormode=%d\n", g_colormode);
		break;  
	case 'w':
		g_wiremode = !g_wiremode;
		break;
	}
	glutPostRedisplay();
}

void reshape(int w, int h)
{
	glViewport(0, 0, w, h);
}

void cleanup()
{
	spMain.deleteProgram();
}

int main( int argc, char **argv )
{
	glutInit(&argc, argv);
    glutInitDisplayMode( GLUT_RGBA );
    glutInitWindowSize( 512, 512 );
    glutCreateWindow(argv[0]);
    
	glewInit();
    init();
    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
	glutReshapeFunc( reshape );
	glutMainLoop();

	cleanup();
    return 0;
}
